/*
 * magnify - a X utility for magnifying screen image with real-time updating
 * 
 * 2000/5/21  Alfredo K. Kojima
 * 
 * This program is in the Public Domain.
 */

#include <X11/Xproto.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/select.h>


int refreshrate = 200;


typedef struct {
    XRectangle *rects;
    int rectP;
    unsigned long lastpixel;
    unsigned long *buffer;
    int rwidth, rheight;	       /* real size */
    int width, height;		       /* size of magnified */
    int magfactor;

    Window win;
    Pixmap pixmap;

    int x, y;
    Bool frozen;
    Bool closed;
    Bool firstdraw;
} BufferData;




static BufferData *newWindow(int magfactor);


int windowCount = 0;

int rectBufferSize = 32;
Display *dpy;
unsigned int black;

BufferData **buffers;
int lastBuffer = 0;

Atom WM_DELETE_WINDOW;


#ifndef __GNUC__
#define inline
#endif



void *wmalloc(unsigned size)
{
    void *ptr = malloc(size);

    if (ptr == NULL) {
	puts("out of memory");
	exit(1);
    }
    return ptr;
}


void *wrealloc(void *data, unsigned size)
{
    void *ptr = realloc(data, size);

    if (ptr == NULL) {
	puts("out of memory");
	exit(1);
    }
    return ptr;
}





static BufferData*
makeBufferData(Window win, int width, int height, int magfactor)
{
    BufferData *data;
    
    
    data = wmalloc(sizeof(BufferData));
    
    data->rwidth = width;
    data->rheight = height;
    
    data->magfactor = magfactor;
    
    data->rects = wmalloc(sizeof(XRectangle)*rectBufferSize);
    data->rectP = 0;
    
    data->win = win;

    data->pixmap = XCreatePixmap(dpy, DefaultRootWindow(dpy), width, height,
				 DefaultDepth(dpy, DefaultScreen(dpy)));
    
    data->frozen = False;
    data->closed = False;

    width /= magfactor;
    height /= magfactor;
    data->buffer = wmalloc(sizeof(unsigned long)*width*height);
    memset(data->buffer, 0, width*height*sizeof(unsigned long));
    data->width = width;
    data->height = height;
    
    data->firstdraw = True;

    XSetWindowBackgroundPixmap(dpy, data->win, data->pixmap);

    return data;
}


static void
resizeBufferData(BufferData *data, int width, int height, int magfactor)
{
    int w = width/magfactor;
    int h = height/magfactor;

    data->rwidth = width;
    data->rheight = height;
    
    data->magfactor = magfactor;
    data->buffer = wrealloc(data->buffer, sizeof(unsigned long)*w*h);
    data->width = w;
    data->height = h;
    memset(data->buffer, 0, w*h*sizeof(unsigned long));
    
    data->firstdraw = True;

    XFreePixmap(dpy, data->pixmap);
    data->pixmap = XCreatePixmap(dpy, DefaultRootWindow(dpy), width, height,
				 DefaultDepth(dpy, DefaultScreen(dpy)));
    
    XSetWindowBackgroundPixmap(dpy, data->win, data->pixmap);
}


static int
drawpoint(BufferData *data, unsigned long pixel, int x, int y)
{
    static GC gc = NULL;

    if (data->buffer[x+data->width*y] == pixel && !data->firstdraw)
	return 0;
    
    data->buffer[x+data->width*y] = pixel;

    if (gc == NULL) {
	XGCValues gcv;
	gcv.graphics_exposures = False;
	gc = XCreateGC(dpy, DefaultRootWindow(dpy), 
		       GCGraphicsExposures, &gcv);
    }
    
    if (data->lastpixel == pixel && data->rectP < rectBufferSize) {
	data->rects[data->rectP].x = x*data->magfactor;
	data->rects[data->rectP].y = y*data->magfactor;
	data->rects[data->rectP].width = data->magfactor;
	data->rects[data->rectP].height = data->magfactor;
	data->rectP++;

	return 0;
    }
    XSetForeground(dpy, gc, data->lastpixel);
    XFillRectangles(dpy, data->pixmap, gc, data->rects, data->rectP);
    data->rectP = 0;
    data->rects[data->rectP].x = x*data->magfactor;
    data->rects[data->rectP].y = y*data->magfactor;
    data->rects[data->rectP].width = data->magfactor;
    data->rects[data->rectP].height = data->magfactor;
    data->rectP++;
   
    data->lastpixel = pixel;

    return 1;
}


static inline unsigned long
getpix(XImage *image, int x, int y, int xoffs, int yoffs)
{
    if (x < xoffs || y < yoffs
	|| x >= xoffs + image->width || y >= yoffs + image->height) {
	return black;
    }
    return XGetPixel(image, x-xoffs, y-yoffs);
}


static void
updateImage(BufferData *data, int rx, int ry)
{
    int gx, gy, gw, gh;
    int x, y;
    int xoffs, yoffs;
    int changedPixels = 0;
    XImage *image;

    gw = data->width;
    gh = data->height;
    
    gx = rx - gw/2;
    gy = ry - gh/2;
    
    xoffs = yoffs = 0;
    if (gx < 0) {
	xoffs = abs(gx);
	gw += gx;
	gx = 0;
    }
    if (gx + gw >= WidthOfScreen(DefaultScreenOfDisplay(dpy))) {
	gw = WidthOfScreen(DefaultScreenOfDisplay(dpy)) - gx;
    }
    if (gy < 0) {
	yoffs = abs(gy);
	gh += gy;
	gy = 0;
    }
    if (gy + gh >= HeightOfScreen(DefaultScreenOfDisplay(dpy))) {
	gh = HeightOfScreen(DefaultScreenOfDisplay(dpy)) - gy;
    }

    image = XGetImage(dpy, DefaultRootWindow(dpy), gx, gy, gw, gh,
		      AllPlanes, ZPixmap);


    for (y = 0; y < data->height; y++) {
	for (x = 0; x < data->width; x++) {
	    unsigned long pixel;

	    pixel = getpix(image, x-xoffs, y-yoffs, xoffs, yoffs);

	    if (drawpoint(data, pixel, x, y))
		changedPixels++;
	}
    }
    
    data->firstdraw = False;

    XDestroyImage(image);

    if (changedPixels > 0) {
	XClearWindow(dpy, data->win);
    }
}


static void
update(BufferData *data)
{
    Window win;
    int rx, ry;
    int bla;
    unsigned ubla;
    

    if (data->frozen) {
	rx = data->x;
	ry = data->y;
    } else {
	XQueryPointer(dpy, DefaultRootWindow(dpy), &win, &win, &rx, &ry, 
		      &bla, &bla, &ubla);
    }
    updateImage(data, rx, ry);
}




static void
keyHandler(XEvent *event, BufferData *data)
{
    char buf[32];
    KeySym ks;
    BufferData *nd;

    if (XLookupString(&event->xkey, buf, 31, &ks, NULL) > 0) {
	switch (buf[0]) {
	 case 'n':
	    nd = newWindow(data->magfactor);
	    buffers = wrealloc(buffers, sizeof(BufferData*)*(lastBuffer+1));
	    buffers[lastBuffer++] = nd;
	    break;
	 case 'f':
	 case ' ':
	    data->frozen = !data->frozen;
	    if (data->frozen) {
		data->x = event->xkey.x_root;
		data->y = event->xkey.y_root;
		sprintf(buf, "[Magnify %ix]", data->magfactor);
	    } else {
		sprintf(buf, "Magnify %ix", data->magfactor);
	    }
	    XStoreName(dpy, data->win, buf);
	    break;
	 case '1':
	 case '2':
	 case '3':
	 case '4':
	 case '5':
	 case '6':
	 case '7':
	 case '8':
	 case '9':
	    resizeBufferData(data, data->rwidth, data->rheight, buf[0]-'0');
	    if (data->frozen) {
		sprintf(buf, "[Magnify %ix]", data->magfactor);
	    } else {
		sprintf(buf, "Magnify %ix", data->magfactor);
	    }
	    XStoreName(dpy, data->win, buf);
	    break;
	}
    }
}


static BufferData*
newWindow(int magfactor)
{
    Window win;
    BufferData *data;
    char buf[32];
    XClassHint *hint = XAllocClassHint();
    Atom protocols[1];

    windowCount++;

    win = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy),
			      0, 0, 300, 200,
			      0, 0, 0);

    hint->res_name = "xmagnify";
    hint->res_class = "XMagnify";
    XSetClassHint(dpy, win, hint);
    XFree(hint);
    
    XSelectInput(dpy, win, StructureNotifyMask|KeyReleaseMask);

    sprintf(buf, "Magnify %ix", magfactor);
    XStoreName(dpy, win, buf);

    data = makeBufferData(win, 300, 200, magfactor);

    XMapRaised(dpy, win);

    protocols[0] = WM_DELETE_WINDOW;

    XSetWMProtocols(dpy, win, protocols, 1);

    return data;
}



static Bool
nextEventOrTimeout(XEvent *event, unsigned long millisec)
{
    struct timeval timeout;
    fd_set rset;

    XSync(dpy, False);
    if (XPending(dpy)) {
        XNextEvent(dpy, event);
        return True;
    }
    
    timeout.tv_usec = millisec*1000;
    timeout.tv_sec = timeout.tv_usec/1000000;
    timeout.tv_usec = timeout.tv_usec%1000000;

    FD_ZERO(&rset);
    FD_SET(ConnectionNumber(dpy), &rset);

    if (select(ConnectionNumber(dpy)+1, &rset, NULL, NULL, &timeout) > 0) {
        XNextEvent(dpy, event);
        return True;
    }

    return False;
}




int main(int argc, char **argv)
{
    int i;
    char *display = "";
    int magfactor = 2;
    
    
    for (i = 1; i < argc; i++) {
	if (strcmp(argv[i], "-display")==0) {
	    i++;
	    if (i >= argc)
		goto help;
	    display = argv[i];
	} else if (strcmp(argv[i], "-m")==0) {
	    i++;
	    if (i >= argc)
		goto help;
	    magfactor = atoi(argv[i]);
	    if (magfactor < 1 || magfactor > 32) {
		printf("%s:invalid magnification factor ``%s''\n", argv[0],
		       argv[i]);
		exit(1);
	    }
	} else if (strcmp(argv[i], "-r")==0) {
	    i++;
	    if (i >= argc)
		goto help;
	    refreshrate = atoi(argv[i]);
	    if (refreshrate < 1) {
		printf("%s:invalid refresh rate ``%s''\n", argv[0], argv[i]);
		exit(1);
	    }
	} else if (strcmp(argv[i], "-h")==0 
		   || strcmp(argv[i], "--help")==0) {
	help:
	    printf("Syntax: %s [options]\n",
		   argv[0]);
	    puts("Options:");
	    puts("  -display <display>	display that should be used");
	    puts("  -m <number>		change magnification factor (default 2)");
	    puts("  -r <number>		change refresh delay, in milliseconds (default 200)");
	    puts("Keys:");
	    puts("  1,2,3,4,5,6,7,8,9	change the magnification factor");
	    puts("  <space>, f		freeze the 'camera', making it magnify only the current\n"
		 "			position");
	    puts("  n			create a new window");
	    exit(0);
	}
    }

    dpy = XOpenDisplay("");
    if (!dpy) {
	puts("couldnt open display");
	exit(1);
    }

    WM_DELETE_WINDOW = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
    
    /* calculate how many rectangles we can send in a trip to the server */
    rectBufferSize = XMaxRequestSize(dpy) - 128;
    rectBufferSize /= sizeof(XRectangle);
    if (rectBufferSize < 1)
	rectBufferSize = 1;

    black = BlackPixel(dpy, DefaultScreen(dpy));

    buffers = wmalloc(sizeof(BufferData*)*1);
    buffers[0] = newWindow(magfactor);
    lastBuffer = 1;

    while (1) {
	XEvent ev;
	if (nextEventOrTimeout(&ev, refreshrate)) {
	    for (i = 0; i < lastBuffer; i++) {
		if (buffers[i]->win == ev.xany.window)
		    break;
	    }
	    if (i < lastBuffer) {
		switch (ev.type) {
		 case ConfigureNotify:
		    if (buffers[i]->rwidth != ev.xconfigure.width
			|| buffers[i]->rheight != ev.xconfigure.height) {
			resizeBufferData(buffers[i],
					 ev.xconfigure.width, 
					 ev.xconfigure.height, 
					 buffers[i]->magfactor);
		    }
		    break;

		 case KeyRelease:
		    keyHandler(&ev, buffers[i]);
		    break;

		 case ClientMessage:
		    windowCount--;
		    if (windowCount == 0) {
			exit(0);
		    } else {
			XDestroyWindow(dpy, buffers[i]->win);
			free(buffers[i]->buffer);
			free(buffers[i]->rects);
			XFreePixmap(dpy, buffers[i]->pixmap);
			free(buffers[i]);
		    
			if (i < windowCount-1) {
			    memmove(&buffers[i], &buffers[i+1], windowCount-i);
			}
		    }
		    break;
		}
	    }
	} else {
	    for (i = 0; i < lastBuffer; i++) {
		update(buffers[i]);
	    }
	}
    }

    return 0;
}

